pub mod faker;
pub mod game;
pub mod packet;
pub mod player;
pub mod protocol;
pub mod server;

use std::{
  collections::HashMap,
  io::Cursor,
  net::{IpAddr, Ipv4Addr, SocketAddr, ToSocketAddrs},
  path::PathBuf,
  sync::{Arc, LazyLock, OnceLock},
};

use clap::Parser;
use faker::Faker;
use packet::{rw::read_packet, Packet};
use player::PlayerActor;
use server::{web_main, Manager};
use tokio::{
  net::{TcpListener, UdpSocket},
  sync::RwLock,
};
use tracing::{error, info, Level};
use tracing_subscriber::{filter::FilterFn, layer::SubscriberExt, util::SubscriberInitExt, Layer};
use xtra::{prelude::MessageChannel, Address, Mailbox};

type PacketChannel = MessageChannel<Packet, (), xtra::refcount::Strong>;
pub fn clients() -> &'static RwLock<HashMap<u128, (PacketChannel, SocketAddr)>> {
  static CLIENTS: LazyLock<RwLock<HashMap<u128, (PacketChannel, SocketAddr)>>> = LazyLock::new(|| RwLock::default());

  &CLIENTS
}

static MANAGER: OnceLock<Address<Manager>> = OnceLock::new();
pub fn manager() -> &'static Address<Manager> {
  &MANAGER.get().unwrap()
}

#[derive(clap::Parser)]
#[command(about = "A game server for Super Mario Odyssey Online")]
struct Arguments {
  #[arg(short, long, default_value_t = 1027)]
  tcp_port: u16,
  #[arg(short, long, default_value_t = 1027)]
  udp_port: u16,
  #[arg(short, long, requires = "prox_cert", requires = "prox_key")]
  prox_port: Option<u16>,
  #[arg(long)]
  prox_cert: Option<PathBuf>,
  #[arg(long)]
  prox_key: Option<PathBuf>,
  #[arg(long)]
  enable_faker: bool,
}

#[tokio::main]
async fn main() {
  tracing_subscriber::registry()
    .with(tracing_subscriber::fmt::layer().with_filter(FilterFn::new(|meta| {
      *meta.level() < Level::INFO || meta.module_path().unwrap().contains("smo_server")
    })))
    .init();
  let arguments = Arguments::parse();

  MANAGER.set(web_main(arguments.prox_port, arguments.prox_cert, arguments.prox_key).await).unwrap();

  let tcp = TcpListener::bind(SocketAddr::new(IpAddr::V4(Ipv4Addr::UNSPECIFIED), arguments.tcp_port))
    .await
    .expect("failed to bind tcp listener");

  if arguments.enable_faker {
    // spawn test bot
    info!("spawning faker for testing!");
    let (address, mailbox) = Mailbox::unbounded();
    let address = xtra::spawn_tokio(
      Faker {
        address: address.clone(),
      },
      (address, mailbox),
    );
    clients().write().await.insert(
      1,
      (
        MessageChannel::new(address),
        "0.0.0.0:0".to_socket_addrs().unwrap().next().unwrap(),
      ),
    );
  }

  let socket = Arc::new(
    UdpSocket::bind(SocketAddr::new(IpAddr::V4(Ipv4Addr::UNSPECIFIED), arguments.udp_port))
      .await
      .expect("failed to bind udp socket"),
  );
  tokio::spawn(udp_server(socket.clone()));

  info!(
    "listening on tcp port {}, udp port {}",
    arguments.tcp_port, arguments.udp_port
  );
  loop {
    match tcp.accept().await {
      Ok((stream, addr)) => {
        PlayerActor::new_connection(stream, addr, socket.clone());
      }
      Err(error) => {
        error!("failed to handle connection: {error:?}");
      }
    }
  }
}

async fn udp_server(socket: Arc<UdpSocket>) {
  loop {
    let mut packet = [0u8; 256];
    let (size, _addr) = socket.recv_from(&mut packet).await.unwrap();
      tracing::trace!("got packet from {_addr:?}");
    match read_packet(&mut Cursor::new(&mut packet[..size]), true).await {
      Ok(packet) => {
        // info!("packet from udp {packet:?}");
        if let Some(client) = clients().read().await.get(&packet.user_id) {
          let _ = client.0.send(packet).detach().await;
        }
      }
      Err(error) => {
        error!("udp packet error: {error:?}");
      }
    };
  }
}
